{ Ein Label mit zusätzlicher Rotation }
{ 17.04.2017 Siro }
{ Drehen bisher nur von 0..90 Grad }
{ darüber gibts noch einige Problemchen :-) }

unit Laz_RotateLabel;

interface

uses
  Classes,
  Graphics,
  StdCtrls;

Type TRotateLabel = class(TCustomLabel)
    protected
      procedure CalculatePreferredSize(var PreferredWidth, PreferredHeight: integer;
        WithThemeSpace: Boolean); override;
      procedure DoMeasureTextPosition(var TextTop, TextLeft: integer); override;
    public
      procedure Paint; override;
    published
      property Font;
      property Caption;
      property AutoSize;
  end;

procedure Register;

implementation

uses
  Themes;

{-------------------------------------------------------------------------}
{ liefert das äussere umschliessende Rechteck eines Punkte-Arrays zurück }
Procedure GetPointsRect(var Points: array of TPoint; out res:Trect);
var i:Integer;  MinX,MaxX,MinY,MaxY:Integer;
begin
  // !! nicht mit 0 initialisieren
  // wir initialisieren auf die Koordinaten des ersten Punktes
  MinX:=Points[0].x;
  MaxX:=MinX;
  MinY:=Points[0].y;
  MaxY:=MinY;

  // index 0 braucht man eigentlich nicht abfragen ....
  for i:=1 to High(Points) do begin
    if Points[i].x < MinX then MinX:=Points[i].x;
    if Points[i].x > MaxX then MaxX:=Points[i].x;
    if Points[i].y < MinY then MinY:=Points[i].y;
    if Points[i].y > MaxY then MaxY:=Points[i].y;
  end;

  res:=Rect(MinX,MinY,MaxX,MaxY);
end;
{------------------------------------------------------------------------------}
procedure RotatePoints(var Points    : array of TPoint;  // Punkte Array
                       RotationPoint : TPoint;           // Punkt der Rotation
                       Angle         : Double);          // Drehwinkel in Grad 0..360 im Uhrzeigersinn
var i:Integer;
var p:TPoint;  { !!! Zwischenvariable zwingend erforderlich }
begin
  Angle := -Angle * pi / 180;       // Winkel Grad nach Rad(Bogenmaß) wandeln

  for i:=0 to High(Points) do begin    // alle Polygonpunkte durchlaufen

    { neue X Kooordinate des Punktes berechnen }
    p.x:=round( RotationPoint.x                              // Original Position
             + (Points[i].x - RotationPoint.x) * cos(Angle)  // + deltaX * cos(w)
             - (Points[i].y - RotationPoint.y) * sin(Angle)  // - deltaY * sin(w)
              );

    { neue Y Kooordinate des Punktes berechnen }
    p.y:=round( RotationPoint.y                              // Original Position
             + (Points[i].x - RotationPoint.x) * sin(Angle)  // + deltaX * sin(w)
             + (Points[i].y - RotationPoint.y) * cos(Angle)  // + deltaY * cos(w)
              );

    { ! erst jetzt die neuen beiden Werte für x und y übernehmen }
    Points[i].x:=p.x;
    Points[i].y:=p.y;
  end;
end;

procedure TRotateLabel.CalculatePreferredSize(
  var PreferredWidth, PreferredHeight: integer; WithThemeSpace: Boolean);
var
  R: TRect;
  P: Array[0..3] of TPoint;
begin
  if (Parent = nil) or (not Parent.HandleAllocated) then
    exit;

  inherited CalculatePreferredSize(PreferredWidth, PreferredHeight, WithThemeSpace);

  P[0] := Point(0, 0);
  P[1] := Point(0, PreferredHeight);
  P[2] := Point(PreferredWidth, PreferredHeight);
  P[3] := Point(PreferredWidth, 0);

  RotatePoints(P, P[0], Font.Orientation/10);
  GetPointsRect(P, R);
  PreferredWidth := abs(R.Right - R.Left);
  PreferredHeight := abs(R.Bottom - R.Top);
end;

procedure TRotateLabel.DoMeasureTextPosition(var TextTop: integer;
  var TextLeft: integer);
var
  lTextHeight: integer = 0;
  lTextWidth: integer = 0;
  P: array[0..3] of TPoint;
  R: TRect;
begin
  inherited;

  if (Font.Orientation = 0) then
    exit;

  // Ungedrehte Textgröße (inkl Zeilenumbruch)
  CalculateSize(Width, lTextWidth, lTextHeight);

  P[0] := Point(0, 0);
  P[1] := Point(0, lTextHeight);
  P[2] := Point(lTextWidth, lTextHeight);
  P[3] := Point(lTextWidth, 0);
  RotatePoints(P, P[0], Font.Orientation/10);
  GetPointsRect(P, R);

  TextTop := TextTop - R.Top;
  TextLeft := TextLeft - R.Left;
end;

// 1:1 Kopie der geerbten Methode. Nur ist das Clipping des Canvas deaktiviert.
// Leider kommt man sonst da nicht ran...
procedure TRotateLabel.Paint;
var
  TR : TTextStyle;
  R : TRect;
  TextLeft, TextTop: integer;
  LabelText: string;
  OldFontColor: TColor;
begin
  R := Rect(0,0,Width,Height);
  with Canvas do
  begin
    Brush.Color := Self.Color;
    if (Color<>clNone) and not Transparent then
    begin
      Brush.Style:=bsSolid;
      FillRect(R);
    end;
    Brush.Style:=bsClear;
    Font := Self.Font;

    FillChar(TR,SizeOf(TR),0);
    with TR do
    begin
//      Alignment := BidiFlipAlignment(Self.Alignment, UseRightToLeftAlignment);
      Layout := Self.Layout;
      Opaque := (Color<>clNone) and not Transparent;
      WordBreak := wordWrap;
      SingleLine:= not WordWrap and not HasMultiLine;
  //    Clipping := True;   // !!!!!!!!!!!!!! DAS MUSS WEG !!!!!!!!!!!!
      ShowPrefix := ShowAccelChar;
      SystemFont := False;
      RightToLeft := UseRightToLeftReading;
      ExpandTabs := True;
    end;
    DoMeasureTextPosition(TextTop, TextLeft);
    LabelText := GetLabelText;
    OldFontColor := Font.Color;
    if not IsEnabled then
      if ThemeServices.ThemesEnabled then
        Font.Color := clGrayText
      else
      begin
        Font.Color := clBtnHighlight;
        TextRect(R, TextLeft + 1, TextTop + 1, LabelText, TR);
        Font.Color := clBtnShadow;
      end;
    TextRect(R, TextLeft, TextTop, LabelText, TR);
    Font.Color := OldFontColor;
  end;
end;

procedure Register;
begin
  RegisterComponents('LAZ_Siro', [TRotateLabel]);
end;

end.

